"""
Unit and regression test for the cmselemental package.
"""

# Import package, test suite, and other packages as needed
import cmselemental
import sys


def test_cmselemental_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "cmselemental" in sys.modules
